import itertools
import time

import pytest
import numpy
from scipy.special import iv

from helpers import *
from .cbrng_ref import philox as philox_ref
from .cbrng_ref import threefry as threefry_ref

from reikna.core import Type
from reikna.helpers import product
from reikna.cbrng import CBRNG
from reikna.cbrng.bijections import threefry, philox
from reikna.cbrng.tools import KeyGenerator
from reikna.cbrng.samplers import uniform_integer, uniform_float, normal_bm, gamma, vonmises
import reikna.cluda.dtypes as dtypes


def uniform_discrete_mean_and_std(min, max):
    return (min + max) / 2., numpy.sqrt(((max - min + 1) ** 2 - 1.) / 12)


def uniform_mean_and_std(min, max):
    return (min + max) / 2., (max - min) / numpy.sqrt(12)


class UniformIntegerHelper:
    def __init__(self, min_, max_):
        self.extent = (min_, max_)
        self.mean, self.std = uniform_discrete_mean_and_std(*self.extent)
        self.name = 'uniform_integer'

    def get_sampler(self, bijection, dtype):
        return uniform_integer(bijection, dtype, self.extent[0], self.extent[1] + 1)

class UniformFloatHelper:
    def __init__(self, min_, max_):
        self.extent = (min_, max_)
        self.mean, self.std = uniform_mean_and_std(*self.extent)
        self.name = 'uniform_float'

    def get_sampler(self, bijection, double):
        dtype = numpy.float64 if double else numpy.float32
        return uniform_float(bijection, dtype, self.extent[0], self.extent[1])

class NormalBMHelper:
    def __init__(self, mean, std):
        self.extent = None
        self.mean = mean
        self.std = std
        self.name = 'normal_bm'

    def get_sampler(self, bijection, double):
        dtype = numpy.float64 if double else numpy.float32
        return normal_bm(bijection, dtype, mean=self.mean, std=self.std)

class NormalBMComplexHelper:
    def __init__(self, mean, std):
        self.extent = None
        self.mean = mean
        self.std = std
        self.name = 'normal_bm_complex'

    def get_sampler(self, bijection, double):
        dtype = numpy.complex128 if double else numpy.complex64
        return normal_bm(bijection, dtype, mean=self.mean, std=self.std)

class GammaHelper:
    def __init__(self, shape, scale):
        self._shape = shape
        self._scale = scale
        self.mean = shape * scale
        self.std = numpy.sqrt(shape) * scale
        self.extent = None
        self.name = 'gamma'

    def get_sampler(self, bijection, double):
        dtype = numpy.float64 if double else numpy.float32
        return gamma(bijection, dtype, shape=self._shape, scale=self._scale)


class VonMisesHelper:
    def __init__(self, mu, kappa):
        self._mu = mu
        self._kappa = kappa
        self.circular_mean = mu
        self.circular_var = 1 - iv(1, kappa) / iv(0, kappa)
        self.extent = (-numpy.pi, numpy.pi)
        self.name = 'vonmises'

    def get_sampler(self, bijection, double):
        dtype = numpy.float64 if double else numpy.float32
        return vonmises(bijection, dtype, mu=self._mu, kappa=self._kappa)


class BijectionHelper:

    def __init__(self, name, words, bitness):
        rounds = 20 if name == 'threefry' else 10
        if name == 'philox':
            bijection_func = philox
        else:
            bijection_func = threefry
        self._name = name
        self._words = words
        self._bitness = bitness
        self._rounds = rounds
        self.bijection = bijection_func(bitness, words, rounds=rounds)

        func = philox_ref if name == 'philox' else threefry_ref
        self._reference_func = lambda ctr, key: func(bitness, words, ctr, key, Nrounds=rounds)

    def reference(self, counters, keygen):
        result = numpy.empty_like(counters)
        for i in range(counters.shape[0]):
            result[i]['v'] = self._reference_func(counters[i]['v'], keygen(i)['v'])
        return result

    def __str__(self):
        return "{name}-{words}x{bitness}-{rounds}".format(
            name=self._name, words=self._words, bitness=self._bitness, rounds=self._rounds)


def pytest_generate_tests(metafunc):

    if 'test_bijection' in metafunc.fixturenames:

        vals = []
        ids = []

        for name, words, bitness in itertools.product(['threefry', 'philox'], [2, 4], [32, 64]):
            val = BijectionHelper(name, words, bitness)
            vals.append(val)
            ids.append(str(val))

        metafunc.parametrize('test_bijection', vals, ids=ids)

    if 'test_sampler_int' in metafunc.fixturenames:
        vals = [
            UniformIntegerHelper(-10, 98)]
        ids = [test.name for test in vals]
        metafunc.parametrize('test_sampler_int', vals, ids=ids)

    if 'test_sampler_float' in metafunc.fixturenames:
        vals = [
            UniformFloatHelper(-5, 7.7),
            NormalBMHelper(-2, 10),
            NormalBMComplexHelper(-3 + 4j, 7),
            GammaHelper(3, 10),
            VonMisesHelper(1, 0.7),
            ]

        ids = [test.name for test in vals]
        metafunc.parametrize('test_sampler_float', vals, ids=ids)


def test_kernel_bijection(thr, test_bijection):

    size = 1000
    seed = 123

    bijection = test_bijection.bijection
    keygen = KeyGenerator.create(bijection, seed=seed, reserve_id_space=False)
    counters_ref = numpy.zeros(size, bijection.counter_dtype)

    rng_kernel = thr.compile_static(
        """
        KERNEL void test(GLOBAL_MEM ${bijection.module}Counter *dest, int ctr)
        {
            VIRTUAL_SKIP_THREADS;
            const VSIZE_T idx = virtual_global_id(0);

            ${bijection.module}Key key = ${keygen.module}key_from_int(idx);
            ${bijection.module}Counter counter = ${bijection.module}make_counter_from_int(ctr);
            ${bijection.module}Counter result = ${bijection.module}bijection(key, counter);

            dest[idx] = result;
        }
        """,
        'test', size,
        render_kwds=dict(bijection=bijection, keygen=keygen))

    dest = thr.array(size, bijection.counter_dtype)

    rng_kernel(dest, numpy.int32(0))
    dest_ref = test_bijection.reference(counters_ref, keygen.reference)
    assert (dest.get() == dest_ref).all()

    rng_kernel(dest, numpy.int32(1))
    counters_ref['v'][:,-1] = 1
    dest_ref = test_bijection.reference(counters_ref, keygen.reference)
    assert (dest.get() == dest_ref).all()


def check_kernel_sampler(thr, sampler, ref):

    size = 10000
    batch = 100
    seed = 456

    bijection = sampler.bijection
    keygen = KeyGenerator.create(bijection, seed=seed)

    rng_kernel = thr.compile_static(
        """
        KERNEL void test(GLOBAL_MEM ${ctype} *dest, int ctr_start)
        {
            VIRTUAL_SKIP_THREADS;
            const VSIZE_T idx = virtual_global_id(0);

            ${bijection.module}Key key = ${keygen.module}key_from_int(idx);
            ${bijection.module}Counter ctr = ${bijection.module}make_counter_from_int(ctr_start);
            ${bijection.module}State st = ${bijection.module}make_state(key, ctr);

            ${sampler.module}Result res;

            for(int j = 0; j < ${batch}; j++)
            {
                res = ${sampler.module}sample(&st);

                %for i in range(sampler.randoms_per_call):
                dest[j * ${size * sampler.randoms_per_call} + ${size * i} + idx] = res.v[${i}];
                %endfor
            }
        }
        """,
        'test', size,
        render_kwds=dict(
            size=size, batch=batch, ctype=dtypes.ctype(sampler.dtype),
            bijection=bijection, keygen=keygen, sampler=sampler))

    dest = thr.array((batch, sampler.randoms_per_call, size), sampler.dtype)
    rng_kernel(dest, numpy.int32(0))
    dest = dest.get()

    check_distribution(dest, ref)


def check_distribution(arr, ref):
    extent = getattr(ref, 'extent', None)
    mean = getattr(ref, 'mean', None)
    std = getattr(ref, 'std', None)
    circular_mean = getattr(ref, 'circular_mean', None)
    circular_var = getattr(ref, 'circular_var', None)

    if extent is not None:
        assert arr.min() >= extent[0]
        assert arr.max() <= extent[1]

    if circular_mean is not None and circular_var is not None:
        z = numpy.exp(1j * arr)
        arr_cmean = numpy.angle(z.mean())
        arr_R = numpy.abs(z.mean())
        arr_cvar = 1 - arr_R

        # FIXME: need a valid mathematical formula for the standard error of the mean
        # for circular distributions.
        # Currently it is just a rough estimate.
        m_std = circular_var**0.5 / numpy.sqrt(arr.size)
        diff = abs(arr_cmean - circular_mean)
        assert diff < 5 * m_std

    if mean is not None and std is not None:
        # expected mean and std of the mean of the sample array
        m_mean = mean
        m_std = std / numpy.sqrt(arr.size)

        diff = abs(arr.mean() - mean)
        assert diff < 5 * m_std # about 1e-6 chance of fail

    if std is not None:
        # expected mean and std of the variance of the sample array
        v_mean = std ** 2
        v_std = numpy.sqrt(2. * std ** 4 / (arr.size - 1))

        diff = abs(arr.var() - v_mean)
        assert diff < 5 * v_std # about 1e-6 chance of fail


def test_32_to_64_bit(thr):
    bijection = philox(32, 4)
    ref = UniformIntegerHelper(0, 2**63-1)
    sampler = ref.get_sampler(bijection, numpy.uint64)
    check_kernel_sampler(thr, sampler, ref)


def test_64_to_32_bit(thr):
    bijection = philox(64, 4)
    ref = UniformIntegerHelper(0, 2**31-1)
    sampler = ref.get_sampler(bijection, numpy.uint32)
    check_kernel_sampler(thr, sampler, ref)


def test_kernel_sampler_int(thr, test_sampler_int):
    bijection = philox(64, 4)
    check_kernel_sampler(
        thr, test_sampler_int.get_sampler(bijection, numpy.int32), test_sampler_int)

def test_kernel_sampler_float(thr_and_double, test_sampler_float):
    thr, double = thr_and_double
    bijection = philox(64, 4)
    check_kernel_sampler(
        thr, test_sampler_float.get_sampler(bijection, double), test_sampler_float)


def check_computation(thr, rng, ref):
    dest_dev = thr.empty_like(rng.parameter.randoms)
    counters = rng.create_counters()
    counters_dev = thr.to_device(counters)
    rngc = rng.compile(thr)

    rngc(counters_dev, dest_dev)
    dest = dest_dev.get()
    check_distribution(dest, ref)


def test_computation_general(thr_and_double):

    size = 10000
    batch = 101

    thr, double = thr_and_double

    bijection = philox(64, 4)
    ref = NormalBMHelper(mean=-2, std=10)
    sampler = ref.get_sampler(bijection, double)

    rng = CBRNG(Type(sampler.dtype, shape=(batch, size)), 1, sampler)
    check_computation(thr, rng, ref)


def test_computation_convenience(thr):

    size = 10000
    batch = 101

    ref = UniformIntegerHelper(0, 511)
    rng = CBRNG.uniform_integer(Type(numpy.int32, shape=(batch, size)), 1,
        sampler_kwds=dict(low=ref.extent[0], high=ref.extent[1]))
    check_computation(thr, rng, ref)


def test_computation_uniqueness(thr):
    """
    A regression test for the bug with a non-updating counter.
    """

    size = 10000
    batch = 1

    rng = CBRNG.normal_bm(Type(numpy.complex64, shape=(batch, size)), 1)

    dest1_dev = thr.empty_like(rng.parameter.randoms)
    dest2_dev = thr.empty_like(rng.parameter.randoms)
    counters = rng.create_counters()
    counters_dev = thr.to_device(counters)
    rngc = rng.compile(thr)

    rngc(counters_dev, dest1_dev)
    rngc(counters_dev, dest2_dev)

    assert not diff_is_negligible(dest1_dev.get(), dest2_dev.get(), verbose=False)


@pytest.mark.perf
@pytest.mark.returns('GB/s')
def test_computation_performance(thr_and_double, fast_math, test_sampler_float):

    thr, double = thr_and_double

    size = 2 ** 15
    batch = 2 ** 6

    bijection = philox(64, 4)
    sampler = test_sampler_float.get_sampler(bijection, double)

    rng = CBRNG(Type(sampler.dtype, shape=(batch, size)), 1, sampler)

    dest_dev = thr.empty_like(rng.parameter.randoms)
    counters = rng.create_counters()
    counters_dev = thr.to_device(counters)
    rngc = rng.compile(thr, fast_math=fast_math)

    attempts = 10
    times = []
    for i in range(attempts):
        t1 = time.time()
        rngc(counters_dev, dest_dev)
        thr.synchronize()
        times.append(time.time() - t1)

    byte_size = size * batch * sampler.dtype.itemsize
    return min(times), byte_size



"""
General purpose algorithms.


Pure parallel computations
^^^^^^^^^^^^^^^^^^^^^^^^^^

.. autoclass:: PureParallel
    :members:


Transposition (permutation)
^^^^^^^^^^^^^^^^^^^^^^^^^^^

.. autoclass:: Transpose
    :members:


Reduction
^^^^^^^^^

.. autoclass:: Reduce
    :members:

Scan
^^^^

.. autoclass:: Scan
    :members:


Predicates
^^^^^^^^^^

.. autoclass:: Predicate
    :members:

.. autofunction:: predicate_sum
"""

from reikna.algorithms.pureparallel import PureParallel
from reikna.algorithms.transpose import Transpose
from reikna.algorithms.reduce import Reduce
from reikna.algorithms.scan import Scan
from reikna.algorithms.predicates import Predicate, predicate_sum



import numpy

from reikna.cluda import Snippet


class Predicate:
    """
    A predicate used in some of Reikna algorithms
    (e.g. :py:class:`~reikna.algorithms.Reduce` or :py:class:`~reikna.algorithms.Scan`).

    :param operation: a :py:class:`~reikna.cluda.Snippet` object with two parameters
        which will take the names of two arguments to join.
    :param empty: a numpy scalar with the empty value of the argument
        (the one which, being joined by another argument, does not change it).
    """

    def __init__(self, operation, empty):
        self.operation = operation
        self.empty = empty

    def __process_modules__(self, process):
        return Predicate(process(self.operation), self.empty)


def predicate_sum(dtype):
    """
    Returns a :py:class:`~reikna.algorithms.Predicate` object which sums its arguments.
    """
    return Predicate(
        Snippet.create(lambda v1, v2: "return ${v1} + ${v2};"),
        numpy.zeros(1, dtype)[0])



"""
This module contains various auxiliary functions which are used throughout the library.
"""

from __future__ import division

import functools
import itertools
import collections
import sys
if sys.version_info[0] >= 3:
    from collections.abc import Iterable
else:
    from collections import Iterable
import os.path
import warnings
import inspect
import funcsigs

from mako.template import Template


class Graph:

    def __init__(self, pairs=None):
        self._pairs = set()
        self._nodes = collections.defaultdict(set)
        if pairs is not None:
            self.add_edges(pairs)

    def add_edge(self, node1, node2):
        assert node1 != node2
        self._nodes[node1].add(node2)
        self._nodes[node2].add(node1)
        self._pairs.add(tuple(sorted((node1, node2))))

    def add_edges(self, pairs):
        for node1, node2 in pairs:
            self.add_edge(node1, node2)

    def add_graph(self, graph):
        for node1, node2 in graph.pairs():
            self.add_edge(node1, node2)

    def add_cluster(self, nodes):
        self.add_edges(itertools.combinations(nodes, 2))

    def remove_node(self, node):
        deps = self._nodes[node]
        for dep in deps:
            self._nodes[dep].remove(node)
            self._pairs.remove(tuple(sorted((node, dep))))
        del self._nodes[node]

    def remove_edge(self, node1, node2):
        assert node1 != node2
        self._pairs.remove(tuple(sorted((node1, node2))))

        self._nodes[node1].remove(node2)
        if len(self._nodes[node1]) == 0:
            del self._nodes[node1]

        self._nodes[node2].remove(node1)
        if len(self._nodes[node2]) == 0:
            del self._nodes[node2]

    def __getitem__(self, node):
        return self._nodes[node]

    def pairs(self):
        return self._pairs

    def translate(self, translator):
        pairs = []
        for node1, node2 in self._pairs:
            pairs.append(tuple(sorted((translator(node1), translator(node2)))))
        return Graph(pairs)


def product(seq):
    """
    Returns the product of elements in the iterable ``seq``.
    """
    return functools.reduce(lambda x1, x2: x1 * x2, seq, 1)


def make_template(template, filename=False):
    # Can't import submodules from reikna itself here, because it creates circular dependencies
    # (computation modules have template_for() calls at the root level,
    # so they get activated on import).
    kwds = dict(
        future_imports=['division'],
        strict_undefined=True,
        imports=['import numpy'])

    # Creating a template from a filename results in more comprehensible stack traces,
    # so we are taking advantage of this if possible.
    if filename:
        kwds['filename'] = template
        return Template(**kwds)
    else:
        return Template(template, **kwds)


def template_from(template):
    """
    Creates a Mako template object from a given string.
    If ``template`` already has ``render()`` method, does nothing.
    """
    if hasattr(template, 'render'):
        return template
    else:
        return make_template(template)


def extract_signature_and_value(func_or_str, default_parameters=None):
    if not inspect.isfunction(func_or_str):
        if default_parameters is None:
            parameters = []
        else:
            kind = funcsigs.Parameter.POSITIONAL_OR_KEYWORD
            parameters = [funcsigs.Parameter(name, kind=kind) for name in default_parameters]

        return funcsigs.Signature(parameters), func_or_str

    signature = funcsigs.signature(func_or_str)

    # pass mock values to extract the value
    args = [None] * len(signature.parameters)
    return signature, func_or_str(*args)


def template_def(signature, code):
    """
    Returns a ``Mako`` template with the given ``signature``.

    :param signature: a list of postitional argument names,
        or a ``Signature`` object from ``funcsigs`` module.
    :code: a body of the template.
    """
    if not isinstance(signature, funcsigs.Signature):
        # treating ``signature`` as a list of positional arguments
        # HACK: Signature or Parameter constructors are not documented.
        kind = funcsigs.Parameter.POSITIONAL_OR_KEYWORD
        signature = funcsigs.Signature([funcsigs.Parameter(name, kind=kind) for name in signature])

    template_src = "<%def name='_func" + str(signature) + "'>\n" + code + "\n</%def>"
    return template_from(template_src).get_def('_func')


def template_for(filename):
    """
    Returns the Mako template object created from the file
    which has the same name as ``filename`` and the extension ``.mako``.
    Typically used in computation modules as ``template_for(__filename__)``.
    """
    name, _ext = os.path.splitext(os.path.abspath(filename))
    return make_template(name + '.mako', filename=True)


def min_blocks(length, block):
    """
    Returns minimum number of blocks with length ``block``
    necessary to cover the array with length ``length``.
    """
    return (length - 1) // block + 1


def log2(num):
    """
    Integer-valued logarigthm with base 2.
    If ``n`` is not a power of 2, the result is rounded to the smallest number.
    """
    pos = 0
    for pow_ in [16, 8, 4, 2, 1]:
        if num >= 2 ** pow_:
            num //= (2 ** pow_)
            pos += pow_
    return pos


def bounding_power_of_2(num):
    """
    Returns the minimal number of the form ``2**m`` such that it is greater or equal to ``n``.
    """
    if num == 1:
        return 1
    else:
        return 2 ** (log2(num - 1) + 1)


def factors(num, limit=None):
    """
    Returns the list of pairs ``(factor, num/factor)`` for all factors of ``num``
    (including 1 and ``num``), sorted by ``factor``.
    If ``limit`` is set, only pairs with ``factor <= limit`` are returned.
    """
    if limit is None or limit > num:
        limit = num

    float_sqrt = num ** 0.5
    int_sqrt = int(round(float_sqrt))

    result = []

    if int_sqrt ** 2 == num:
        int_limit = int_sqrt + 1
    else:
        int_limit = int(float_sqrt) + 1

    for i in range(1, int_limit):
        div, mod = divmod(num, i)
        if mod == 0:
            result.append((i, div))

    if limit > result[-1][0]:
        if int_sqrt ** 2 == num:
            to_rev = result[:-1]
        else:
            to_rev = result

        result = result + [(div, f) for f, div in reversed(to_rev)]

    return [r for r in result if r[0] <= limit]


def wrap_in_tuple(seq_or_elem):
    """
    If ``seq_or_elem`` is a sequence, converts it to a ``tuple``,
    otherwise returns a tuple with a single element ``seq_or_elem``.
    """
    if seq_or_elem is None:
        return tuple()
    elif isinstance(seq_or_elem, str):
        return (seq_or_elem,)
    elif isinstance(seq_or_elem, Iterable):
        return tuple(seq_or_elem)
    else:
        return (seq_or_elem,)


class ignore_integer_overflow():
    """
    Context manager for ignoring integer overflow in numpy operations on scalars
    (not ignored by default because of a bug in numpy).
    """

    def __init__(self):
        self.catch = warnings.catch_warnings()

    def __enter__(self):
        self.catch.__enter__()
        warnings.filterwarnings("ignore", "overflow encountered in uint_scalars")
        warnings.filterwarnings("ignore", "overflow encountered in ulong_scalars")

    def __exit__(self, *args, **kwds):
        self.catch.__exit__(*args, **kwds)


def normalize_axes(ndim, axes):
    """
    Transform an iterable of array axes (which can be negative) or a single axis
    into a tuple of non-negative axes.
    """
    if axes is None:
        axes = tuple(range(ndim))
    else:
        axes = wrap_in_tuple(axes)
        axes = tuple(axis if axis >= 0 else ndim + axis for axis in axes)
        if any(axis < 0 or axis >= ndim for axis in axes):
            raise IndexError("Array index out of range")
    return axes


def are_axes_innermost(ndim, axes):
    inner_axes = list(range(ndim - len(axes), ndim))
    return all(axis == inner_axis for axis, inner_axis in zip(axes, inner_axes))


def make_axes_innermost(ndim, axes):
    """
    Given the total number of array axes and a list of axes in this range,
    produce a transposition plan (suitable e.g. for ``numpy.transpose()``)
    that will move make the given axes innermost (in the order they're given).
    Returns the transposition plan, and the plan to transpose the resulting array back
    to the original axes order.
    """
    orig_order = list(range(ndim))
    outer_axes = [i for i in orig_order if i not in axes]
    transpose_to = outer_axes + list(axes)

    transpose_from = [None] * ndim
    for i, axis in enumerate(transpose_to):
        transpose_from[axis] = i

    return tuple(transpose_to), tuple(transpose_from)


def default_strides(shape, itemsize):
    """
    Return the default strides (corresponding to a contiguous array) for an array
    of shape ``shape`` and elements of size ``itemsize`` bytes.
    """
    return tuple(product(shape[j+1:]) * itemsize for j in range(len(shape)))


def min_buffer_size(shape, itemsize, strides=None, offset=0):
    """
    Return the minimum memory buffer size (in bytes) that can fit
    an array with given parameters, starting at an ``offset`` bytes
    from the beginning of the buffer.
    """
    if strides is None:
        strides = default_strides(shape, itemsize)

    data_size = sum(
        (stride * (l - 1) if stride > 0 else 0)
        for stride, l in zip(strides, shape)) + itemsize
    return data_size + offset


def padded_buffer_parameters(shape, itemsize, pad=0):
    """
    For an array of shape ``shape``, padded from all sizes with ``pad`` elements,
    return a tuple of (strides, offset, size (in bytes) of the required memory buffer),
    which would have to be requested when allocating such an array.
    """
    padded_shape = [l + pad * 2 for l in shape]
    strides = default_strides(padded_shape, itemsize)
    full_size = min_buffer_size(padded_shape, itemsize, strides=strides)
    offset = sum(stride * pad for stride in strides)
    return strides, offset, full_size



"""
This example illustrates how to manually use a temporary array manager (if you must).
"""

import numpy
from reikna.cluda import dtypes, any_api
from reikna.cluda.tempalloc import ZeroOffsetManager


api = any_api()
thr = api.Thread.create()


def demo_array_dependencies():

    # ZeroOffsetManager attempts to pack temporary allocations
    # in a collection of real allocations with minimal total size.
    # All the virtual allocations start at the beginning of the real allocations.

    # Create a manager that will try to minimize the total size of real allocations
    # every time a temporary allocation occurs, or a temporary array is freed.
    # Note that this may involve re-pointing a temporary array to a different part of memory,
    # so all of the data in it is lost.
    temp_manager = ZeroOffsetManager(thr, pack_on_alloc=True, pack_on_free=True)

    # Alternatively one can pass `False` to these keywords and call `.pack()` manually.
    # This can be useful if a lot of allocations are happening in a specific place at once.

    # Create two arrays that do not depend on each other.
    # This means the manager will allocate a single (200, int32) real array,
    # and point both `a1` and `a2` to its beginning.
    a1 = temp_manager.array(100, numpy.int32)
    a2 = temp_manager.array(200, numpy.int32)

    # You can see that the total size of virtual arrays is 1200,
    # but the total size of real arrays is only 800 (the size of the larger array).
    print("Allocated a1 = (100, int32) and a2 = (200, int32)")
    print(temp_manager._statistics())

    # Now we allocate a dependent array.
    # This means that the real memory `a3` points to cannot intersect with that of `a1`.
    # If we could point temporary arrays at any address within real allocations,
    # we could fit it into the second half of the existing real allocation.
    # But `ZeroOffsetManager` cannot do that, so it has to create another allocation.
    a3 = temp_manager.array(100, numpy.int32, dependencies=[a1])

    print("Allocated a3 = (100, int32) depending on a1")
    print(temp_manager._statistics())

    # Now that we deallocated `a1`, `a3` can now fit in the same real allocation as `a2`,
    # so one of the real allocations will be removed.
    del a1

    print("Freed a1")
    print(temp_manager._statistics())


class MyComputation:

    def __init__(self, temp_manager):
        self.temp_array = temp_manager.array(100, numpy.int32)

        # The magic property containing temporary arrays used
        self.__tempalloc__ = [self.temp_array]

    def __call__(self, array1, array2):
        # a sequence of kernel calls using `self.temp_array` to store some intermediate results
        pass


def demo_object_dependencies():

    temp_manager = ZeroOffsetManager(thr, pack_on_alloc=True, pack_on_free=True)

    # A `MyComputation` instance creates a temporary array for internal usage
    comp = MyComputation(temp_manager)

    print("MyComputation created")
    print(temp_manager._statistics())

    # Create another temporary array whose usage does not intersect with `MyComputation` usage.
    # This means that if `comp` is called, the contents of `a1` may be rewritten.
    a1 = temp_manager.array(100, numpy.int32)

    # It is put in the same real allocation as the temporary array of `comp`.
    print("Allocated a1 = (100, int32)")
    print(temp_manager._statistics())

    # Now let's say we want to put the result of `comp` call somewhere.
    # This means we want to make sure it does not occupy the same memory
    # as any of the temporary arrays in `comp`, so we are passing `comp` as a dependency.
    # It will pick up whatever `comp` declared in its `__tempalloc__` attribute.
    result = temp_manager.array(100, numpy.int32, dependencies=[comp])

    # You can see that a new real allocation was created to host the result.
    print("Allocated result = (100, int32)")
    print(temp_manager._statistics())


if __name__ == '__main__':
    demo_array_dependencies()
    demo_object_dependencies()



"""
This example shows how to implement an n-dimensional FFT over arbitrary axes
using a 1D FFT over the innermost dimension and transpositions.

At the moment of the writing, testing shows that the performance of this approach
is generally worse than that of the existing global FFT kernel
(the worst case being a 3D FFT with dimensions 1024x(16, 16, 16) on CUDA,
where the performance is almost halved).

Nevertheless, this computation is preserved as an example in case it is ever needed,
because it will greatly simplify the FFT computation.
"""

import time

import numpy

from reikna.cluda import any_api
from reikna.core import Computation, Parameter, Annotation
from reikna.fft import FFT
from reikna.algorithms import Transpose


class FFTWithTranspose(Computation):

    def __init__(self, arr_t, axes=None):

        Computation.__init__(self, [
            Parameter('output', Annotation(arr_t, 'o')),
            Parameter('input', Annotation(arr_t, 'i')),
            Parameter('inverse', Annotation(numpy.int32), default=0)])

        if axes is None:
            axes = range(len(arr_t.shape))
        self._axes = tuple(sorted(axes))


    def _build_plan(self, plan_factory, device_params, output, input_, inverse):

        plan = plan_factory()

        num_axes = len(input_.shape)

        current_axes = list(range(num_axes))
        current_input = input_

        # Iterate over all the axes we need to FFT over
        for i, initial_axis in enumerate(self._axes):

            # Find out where the target axis is currently located
            current_axis = current_axes.index(initial_axis)

            # If it is not the innermost one, we will transpose the array
            # to bring it to the end.
            if current_axis != len(current_axes) - 1:

                local_axes = list(range(num_axes))

                # The `Transpose` computation is most efficient when we ask it
                # to swap two innermost parts of the axes list, e.g.
                # [0, 1, 2, 3, 4, 5] to [0, 1, 4, 5, 2, 3]
                # This way the transposition requires only one kernel call.

                # Since we do not care where the other axes go as long as the target one
                # becomes the innermost one, it is easy to follow this guideline.

                # That's the transposition that we will need to perform
                # on the current array
                local_axes = (
                    local_axes[:current_axis]
                    + local_axes[current_axis+1:]
                    + [local_axes[current_axis]])

                # That's the corresponding permutation of the original axes
                # (we need to keep track of it)
                current_axes = (
                    current_axes[:current_axis]
                    + current_axes[current_axis+1:]
                    + [current_axes[current_axis]])

                # Transpose the array, saving the result in a temporary buffer
                tr = Transpose(current_input, axes=local_axes)
                temp = plan.temp_array_like(tr.parameter.output)
                plan.computation_call(tr, temp, current_input)

                # Now the target axis is the innermost one
                current_axis = len(current_axes) - 1
                current_input = temp

            # If it is the last FFT to perform, and there is no final transposition required,
            # the FFT output is pointed to the output array instead of a temporary buffer.
            if i == len(self._axes) - 1 and current_axes == list(range(num_axes)):
                current_output = output
            else:
                current_output = current_input

            fft = FFT(current_input, axes=(current_axis,))
            plan.computation_call(fft, current_output, current_input, inverse)

        # If the axes are not in the original order, there is one last transposition required
        if current_axes != list(range(num_axes)):
            pairs = list(zip(current_axes, list(range(num_axes))))
            tr_axes = [local_axis for _, local_axis in sorted(pairs)]
            tr = Transpose(current_output, tr_axes)
            plan.computation_call(tr, output, current_output)

        return plan


if __name__ == '__main__':

    api = any_api()
    thr = api.Thread.create()

    dtype = numpy.complex128

    shape = (1024, 16, 16, 16)
    axes = (1, 2, 3)

    data = numpy.random.normal(size=shape) + 1j * numpy.random.normal(size=shape)
    data = data.astype(dtype)

    fft = FFT(data, axes=axes)
    fftc = fft.compile(thr)

    fft2 = FFTWithTranspose(data, axes=axes)
    fft2c = fft2.compile(thr)

    data_dev = thr.to_device(data)
    res_dev = thr.empty_like(data_dev)

    for comp, tag in [(fftc, "original FFT"), (fft2c, "transposition-based FFT")]:
        attempts = 10
        ts = []
        for i in range(attempts):
            t1 = time.time()
            comp(res_dev, data_dev)
            thr.synchronize()
            t2 = time.time()
            ts.append(t2 - t1)

        fwd_ref = numpy.fft.fftn(data, axes=axes).astype(dtype)
        assert numpy.allclose(res_dev.get(), fwd_ref)

        print(tag, min(ts), "s")



# FIXME: This module is a part of Array functionality, so it is located on CLUDA level,
# but it requires some high-level Reikna functionality (computations and transformations).
# So it is a bit of circular dependency.
# Ideally, this should be moved to computation level, perhaps creating two versions of Array -
# CLUDA level (without __setitem__) and Reikna level (with one).

import numpy

import reikna.cluda.dtypes as dtypes
import reikna.transformations as transformations
import reikna.cluda.functions as functions
from reikna.algorithms import PureParallel
from reikna.core import Type, Parameter, Annotation, Computation


def normalize_value(thr, gpu_array_type, val):
    """
    Transforms a given value (a scalar or an array)
    to a value that can be passed to a kernel.
    Returns a pair (result, is_array), where ``is_array`` is a boolean
    that helps distinguishing between zero-dimensional arrays and scalars.
    """
    if isinstance(val, gpu_array_type):
        return val, True
    elif isinstance(val, numpy.ndarray):
        return thr.to_device(val), True
    else:
        dtype = dtypes.detect_type(val)
        return numpy.asarray(val, dtype), False


def setitem_computation(dest, source, is_array):
    """
    Returns a compiled computation that broadcasts ``source`` to ``dest``,
    where ``dest`` is a GPU array, and ``source`` is either a GPU array or a scalar.
    """
    if is_array:
        source_dt = Type.from_value(source).with_dtype(dest.dtype)
        trf = transformations.copy(source_dt, dest)
        comp = PureParallel.from_trf(trf, guiding_array=trf.output)
        cast_trf = transformations.cast(source, dest.dtype)
        comp.parameter.input.connect(cast_trf, cast_trf.output, src_input=cast_trf.input)
        return comp
    else:
        trf = transformations.broadcast_param(dest)
        return PureParallel.from_trf(trf, guiding_array=trf.output)


def setitem_method(array, index, value):
    # We need it both in ``cuda.Array`` and ``ocl.Array``, hence a standalone function.

    # PyOpenCL and PyCUDA support __setitem__() for some restricted cases,
    # but it is too complicated to determine when it will work,
    # and it is easier to just call our own implementation every time.

    view = array[index]
    value, is_array = normalize_value(array.thread, type(array), value)
    comp = array.thread.get_cached_computation(
        setitem_computation, Type.from_value(view), Type.from_value(value), is_array)
    comp(view, value)


def get_method(array):
    temp = array.thread.array(array.shape, array.dtype)
    comp = array.thread.get_cached_computation(
        setitem_computation, Type.from_value(temp), Type.from_value(array), True)
    comp(temp, array)
    return temp.get()


def is_shape_compatible(template_shape, shape, axis):
    for i in range(len(template_shape)):
        if i != axis and shape[i] != template_shape[i]:
            return False
    return True


def concatenate(arrays, axis=0, out=None):
    """
    Concatenate an iterable of arrays along ``axis`` and write them to ``out``
    (allocating it if it is set to ``None``).

    Works analogously to ``numpy.concatenate()`` (except ``axis=None`` is not supported).
    """

    # TODO: support axis=None.
    # Requires Array.ravel() returnign an Array instead of CUDA/PyOpenCL array.

    if len(arrays) == 0:
        raise ValueError("Need at least one array to concatenate")
    if any(array.dtype != arrays[0].dtype for array in arrays[1:]):
        raise ValueError("Data types of all arrays must be the same")

    dtype = arrays[0].dtype
    thread = arrays[0].thread

    template_shape = arrays[0].shape
    axis = axis % len(template_shape)
    for array in arrays[1:]:
        if not is_shape_compatible(template_shape, array.shape, axis):
            raise ValueError(
                "Shapes are not compatible: " + str(template_shape) + " and " + str(shape))

    out_shape = list(template_shape)
    out_shape[axis] = sum(array.shape[axis] for array in arrays)
    out_shape = tuple(out_shape)

    if out is None:
        out = thread.array(out_shape, dtype)
    else:
        if out.shape != out_shape:
            raise ValueError(
                "Incorrect output shape: expected " + str(out_shape) + ", got " + str(out.shape))
        if out.dtype != dtype:
            raise ValueError(
                "Incorrect output dtype: expected " + str(dtype) + ", got " + str(out.dtype))

    offset = 0
    slices = [slice(None) for i in range(len(out_shape))]
    for array in arrays:
        slices[axis] = slice(offset, offset + array.shape[axis])
        out[tuple(slices)] = array
        offset += array.shape[axis]

    return out


def roll_computation(array, axis):
    return PureParallel(
        [
            Parameter('output', Annotation(array, 'o')),
            Parameter('input', Annotation(array, 'i')),
            Parameter('shift', Annotation(Type(numpy.int32)))],
        """
        <%
            shape = input.shape
        %>
        %for i in range(len(shape)):
            VSIZE_T output_${idxs[i]} =
                %if i == axis:
                ${shift} == 0 ?
                    ${idxs[i]} :
                    ## Since ``shift`` can be negative, and its absolute value greater than
                    ## ``shape[i]``, a double modulo division is necessary
                    ## (the ``%`` operator preserves the sign of the dividend in C).
                    (${idxs[i]} + (${shape[i]} + ${shift} % ${shape[i]})) % ${shape[i]};
                %else:
                ${idxs[i]};
                %endif
        %endfor
        ${output.store_idx}(
            ${", ".join("output_" + name for name in idxs)},
            ${input.load_idx}(${", ".join(idxs)}));
        """,
        guiding_array='input',
        render_kwds=dict(axis=axis))


class RollInplace(Computation):

    def __init__(self, array, axis):
        self._axis = axis
        Computation.__init__(self, [
            Parameter('array', Annotation(array, 'io')),
            Parameter('shift', Annotation(Type(numpy.int32)))])

    def _build_plan(self, plan_factory, device_params, array, shift):
        plan = plan_factory()

        temp = plan.temp_array_like(array)
        plan.computation_call(roll_computation(array, self._axis), temp, array, shift)

        tr = transformations.copy(temp, out_arr_t=array)
        copy_comp = PureParallel.from_trf(tr, guiding_array=tr.output)
        plan.computation_call(copy_comp, array, temp)

        return plan


def roll(array, shift, axis=-1):
    """
    Cyclically shifts elements of ``array`` by ``shift`` positions to the right along ``axis``.
    ``shift`` can be negative (in which case the elements are shifted to the left).
    Elements that are shifted beyond the last position are re-introduced at the first
    (and vice versa).

    Works equivalently to ``numpy.roll`` (except ``axis=None`` is not supported).
    """
    temp = array.thread.array(array.shape, array.dtype)
    axis = axis % len(array.shape)
    comp = array.thread.get_cached_computation(
        roll_computation, Type.from_value(array), axis)
    comp(temp, array, shift)
    return temp


def roll_method(array, shift, axis=-1):
    axis = axis % len(array.shape)
    comp = array.thread.get_cached_computation(
        RollInplace, Type.from_value(array), axis)
    comp(array, shift)



import sys
from tempfile import mkdtemp
import os.path

import pyopencl as cl
import pyopencl.array as clarray

from reikna.helpers import wrap_in_tuple, product, min_buffer_size
import reikna.cluda as cluda
import reikna.cluda.dtypes as dtypes
import reikna.cluda.api as api_base

from reikna.cluda.array_helpers import setitem_method, get_method, roll_method


def get_id():
    return cluda.ocl_id()


def get_platforms():
    return cl.get_platforms()


class Array(clarray.Array):
    """
    A subclass of PyOpenCL ``Array``, with some additional functionality.
    """
    def __init__(
            self, thr, shape, dtype, strides=None, offset=0, nbytes=None,
            allocator=None, data=None, events=None,
            _fast=False, _context=None, _queue=None, _size=None):

        if thr is None:
            thr = Thread(_queue)

        clarray.Array.__init__(
            self, thr._queue, shape, dtype, strides=strides, allocator=allocator,
            data=data, offset=offset, events=events,
            _fast=_fast, _context=thr._context, _queue=thr._queue, _size=_size)
        self.nbytes = nbytes
        self.thread = thr

    def _new_like_me(self, dtype=None, queue=None):
        """
        Called by PyOpenCL to store the results of arithmetic operations
        or when the array is copied, to make an empty array.
        Need to intercept it to preserve the array type.
        The `queue` argument is ignored, we're always using the queue of the thread.
        """
        return (self.thread.empty_like(self)
                if dtype is None
                else self.thread.array(self.shape, dtype))

    def __getitem__(self, index):
        res = clarray.Array.__getitem__(self, index)

        # Let cl.Array calculate the new strides and offset
        return self.thread.array(
            shape=res.shape, dtype=res.dtype, strides=res.strides,
            base=res.base_data,
            offset=res.offset)

    def __setitem__(self, index, value):
        setitem_method(self, index, value)

    def roll(self, shift, axis=-1):
        roll_method(self, shift, axis=axis)

    def get(self, *args, **kwds):
        if self.flags.forc:
            return clarray.Array.get(self, *args, **kwds)
        else:
            assert len(args) == 0 and len(kwds) == 0
            return get_method(self)

    def _tempalloc_update_buffer(self, data):
        self.base_data = data


class Thread(api_base.Thread):

    api = sys.modules[__name__]

    def _process_cqd(self, cqd):
        if isinstance(cqd, cl.Device):
            context = cl.Context(devices=[cqd])
            return context, cl.CommandQueue(context), cqd, False
        elif isinstance(cqd, cl.Context):
            return cqd, cl.CommandQueue(cqd), cqd.devices[0], False
        elif isinstance(cqd, cl.CommandQueue):
            return cqd.context, cqd, cqd.device, False
        else:
            return ValueError("The value provided is not Device, Context or CommandQueue")

    def array(
            self, shape, dtype, strides=None, offset=0, nbytes=None,
            allocator=None, base=None, data=None):

        if allocator is None:
            allocator = self.allocate

        dtype = dtypes.normalize_type(dtype)
        shape = wrap_in_tuple(shape)
        if nbytes is None:
            nbytes = int(min_buffer_size(shape, dtype.itemsize, strides=strides, offset=offset))

        if (offset != 0 or strides is not None) and data is None and base is None:
            data = allocator(nbytes)
        elif base is not None:
            if isinstance(base, Array):
                base = base.base_data
            data = base

        return Array(
            self, shape, dtype, strides=strides, offset=offset,
            allocator=allocator, data=data, nbytes=nbytes)

    def allocate(self, size):
        return cl.Buffer(self._context, cl.mem_flags.READ_WRITE, size=size)

    def _copy_array(self, dest, src):
        dest.set(src, queue=self._queue, async_=self._async)

    def from_device(self, arr, dest=None, async_=False):
        arr_cpu = arr.get(queue=self._queue, ary=dest, async_=async_)
        if dest is None:
            return arr_cpu

    def _copy_array_buffer(self, dest, src, nbytes, src_offset=0, dst_offset=0):
        cl.enqueue_copy(
            self._queue, dest.data, src.data,
            byte_count=nbytes, src_offset=src_offset, dst_offset=dst_offset)

    def synchronize(self):
        self._queue.finish()

    def _compile(self, src, fast_math=False, compiler_options=None, keep=False):
        options = "-cl-mad-enable -cl-fast-relaxed-math" if fast_math else ""
        if compiler_options is not None:
            options += " " + " ".join(compiler_options)

        if keep:
            temp_dir = mkdtemp()
            temp_file_path = os.path.join(temp_dir, 'kernel.cl')

            with open(temp_file_path, 'w') as f:
                f.write(src)

            print("*** compiler output in", temp_dir)

        else:
            temp_dir = None

        return cl.Program(self._context, src).build(options=options, cache_dir=temp_dir)


class DeviceParameters:

    def __init__(self, device):

        self.api_id = get_id()

        self._device = device

        if device.platform.name == 'Apple' and device.type == cl.device_type.CPU:
        # Apple is being funny again.
        # On OSX 10.8.0 it reports the maximum block size as 1024, when it is really 128.
        # Moreover, if local_barrier() is used in the kernel, it becomes 1
            self.max_work_group_size = 1
            self.max_work_item_sizes = [1, 1, 1]
        else:
            self.max_work_group_size = device.max_work_group_size
            self.max_work_item_sizes = device.max_work_item_sizes

        max_size = 2 ** device.address_bits
        self.max_num_groups = [max_size, max_size, max_size]

        if device.type == cl.device_type.CPU:
            # For CPU both values do not make much sense
            self.local_mem_banks = self.max_work_group_size
            self.warp_size = 1
        elif "cl_nv_device_attribute_query" in device.extensions:
            # If NV extensions are available, use them to query info
            self.local_mem_banks = 16 if device.compute_capability_major_nv < 2 else 32
            self.warp_size = device.warp_size_nv
        elif device.vendor == 'NVIDIA':
            # nVidia device, but no extensions.
            # Must be APPLE OpenCL implementation.
            self.local_mem_banks = 16
            self.warp_size = 16
        else:
            # AMD card.
            # Do not know how to query this info, so settle for most probable values.

            self.local_mem_banks = 32

            # An alternative is to query CL_KERNEL_PREFERRED_WORK_GROUP_SIZE_MULTIPLE
            # for some arbitrary kernel.
            self.warp_size = 64

        self.min_mem_coalesce_width = {4: 16, 8: 16, 16: 8}
        self.local_mem_size = device.local_mem_size

        self.compute_units = device.max_compute_units

    def supports_dtype(self, dtype):
        if dtypes.is_double(dtype):
            extensions = self._device.extensions
            return "cl_khr_fp64" in extensions or "cl_amd_fp64" in extensions
        else:
            return True


class Kernel(api_base.Kernel):

    def _get_kernel(self, program, name):
        return getattr(program, name)

    def _fill_attributes(self):
        self.max_work_group_size = self._kernel.get_work_group_info(
            cl.kernel_work_group_info.WORK_GROUP_SIZE, self._thr._device)

    def prepare(self, global_size, local_size=None, local_mem=0):
        # ``local_mem`` is ignored, since it cannot be easily passed to the kernel
        # (a special kernel argument is requred).
        if local_size is None:
            self._local_size = None
        else:
            self._local_size = wrap_in_tuple(local_size)
        self._global_size = wrap_in_tuple(global_size)

    def _prepared_call(self, *args):
        # Passing data, assuming that the kernel knows how to handle the offset and the strides
        args = [x.base_data if isinstance(x, clarray.Array) else x for x in args]
        return self._kernel(self._thr._queue, self._global_size, self._local_size, *args)



import itertools

import pytest

import reikna.cluda as cluda
import reikna.cluda.dtypes as dtypes
import reikna.cluda.functions as functions
from reikna.cluda import tempalloc
from reikna import concatenate, roll
from reikna.helpers import product

from helpers import *
from pytest_threadgen import parametrize_thread_tuple, create_thread_in_tuple


TEST_DTYPES = [
    numpy.int8, numpy.int16, numpy.int32, numpy.int64,
    numpy.uint8, numpy.uint16, numpy.uint32, numpy.uint64,
    numpy.float32, numpy.float64,
    numpy.complex64, numpy.complex128]


@pytest.fixture
def thr_and_global_size(request):
    return create_thread_in_tuple(request)


def pair_thread_with_gs(metafunc, tp):
    global_sizes = [
        (100,), (2000,), (1153,),
        (10, 10), (150, 250), (137, 547),
        (7, 11, 13), (50, 100, 100), (53, 101, 101)]

    rem_ids = []
    vals = []

    for gs in global_sizes:

        # If the thread will not support these limits, skip
        mgs = tp.device_params.max_num_groups
        if len(gs) > len(mgs) or (len(mgs) > 2 and len(gs) > 2 and mgs[2] < gs[2]):
            continue

        rem_ids.append(str(gs))
        vals.append((gs,))

    return [tp] * len(vals), vals, rem_ids


def pytest_generate_tests(metafunc):
    if 'thr_and_global_size' in metafunc.fixturenames:
        parametrize_thread_tuple(metafunc, 'thr_and_global_size', pair_thread_with_gs)


def simple_thread_test(thr):
    shape = (1000,)
    dtype = numpy.float32

    a = get_test_array(shape, dtype)
    a_dev = thr.to_device(a)
    a_back = thr.from_device(a_dev)

    assert diff_is_negligible(a, a_back)


def test_create_new_thread(cluda_api):
    thr = cluda_api.Thread.create()
    simple_thread_test(thr)


def test_array_functionality(thr):
    """
    Checks additional functionality introduced by the generic Array class.
    """
    arr = thr.array(1024, numpy.float32)
    assert arr.thread is thr


def test_array_type_after_binop(thr):
    """
    Checks that a binary operation keeps the Reikna Array type
    and does not reset it to PyOpenCL/PyCUDA array.
    """
    arr = thr.array(1024, numpy.float32)
    arr2 = thr.array(1024, numpy.float32)

    assert type(arr + arr2) == type(arr)


def test_array_copy(thr):
    """
    Checks that .copy() copies the array, keeps the Reikna Array type
    and does not reset it to PyOpenCL/PyCUDA array.
    """
    arr = thr.to_device(get_test_array(1024, numpy.float32))
    arr2 = arr.copy()

    assert type(arr2) == type(arr)
    assert diff_is_negligible(arr.get(), arr2.get())


def test_transfers(thr):
    a = get_test_array(1024, numpy.float32)

    def to_device1(x):
        return thr.to_device(x)
    def to_device2(x):
        y = thr.empty_like(x)
        thr.to_device(x, dest=y)
        return y
    def from_device1(x):
        return x.get()
    def from_device2(x):
        return thr.from_device(x)
    def from_device3(x):
        y = numpy.empty(x.shape, x.dtype)
        thr.from_device(x, dest=y)
        return y
    def from_device4(x):
        y = thr.from_device(x, async_=True)
        thr.synchronize()
        return y
    def from_device5(x):
        y = numpy.empty(x.shape, x.dtype)
        thr.from_device(x, dest=y, async_=True)
        thr.synchronize()
        return y

    to_device = (to_device1, to_device2)
    from_device = (from_device1, from_device2, from_device3, from_device4, from_device5)

    for to_d, from_d in itertools.product(to_device, from_device):
        a_device = to_d(a)
        a_copy = thr.copy_array(a_device)
        a_back = from_d(a_copy)
        assert diff_is_negligible(a, a_back)


@pytest.mark.parametrize(
    "dtype", TEST_DTYPES,
    ids=[dtypes.normalize_type(dtype).name for dtype in TEST_DTYPES])
def test_dtype_support(thr, dtype):
    # Test passes if either thread correctly reports that it does not support given dtype,
    # or it successfully compiles kernel that operates with this dtype.

    N = 256

    if not thr.device_params.supports_dtype(dtype):
        pytest.skip()

    mul = functions.mul(dtype, dtype)
    div = functions.div(dtype, dtype)
    program = thr.compile(
    """
    KERNEL void test(
        GLOBAL_MEM ${ctype} *dest, GLOBAL_MEM ${ctype} *a, GLOBAL_MEM ${ctype} *b)
    {
      const SIZE_T i = get_global_id(0);
      ${ctype} temp = ${mul}(a[i], b[i]);
      dest[i] = ${div}(temp, b[i]);
    }
    """, render_kwds=dict(ctype=dtypes.ctype(dtype), dtype=dtype, mul=mul, div=div))

    test = program.test

    # we need results to fit even in unsigned char
    a = get_test_array(N, dtype, high=8)
    b = get_test_array(N, dtype, no_zeros=True, high=8)

    a_dev = thr.to_device(a)
    b_dev = thr.to_device(b)
    dest_dev = thr.empty_like(a_dev)
    test(dest_dev, a_dev, b_dev, global_size=N)
    assert diff_is_negligible(thr.from_device(dest_dev), a)


def test_find_local_size(thr_and_global_size):
    """
    Check that if None is passed as local_size, kernel can find some local_size to run with
    (not necessarily optimal).
    """

    thr, global_size = thr_and_global_size

    program = thr.compile(
    """
    KERNEL void test(GLOBAL_MEM int *dest)
    {
        const SIZE_T i = get_global_id(0) +
            get_global_id(1) * get_global_size(0) +
            get_global_id(2) * get_global_size(1) * get_global_size(0);
        dest[i] = i;
    }
    """)
    test = program.test
    dest_dev = thr.array(global_size, numpy.int32)
    test(dest_dev, global_size=global_size)

    assert diff_is_negligible(dest_dev.get().ravel(),
        numpy.arange(product(global_size)).astype(numpy.int32))


@pytest.mark.parametrize(
    'tempalloc_cls',
    [tempalloc.TrivialManager, tempalloc.ZeroOffsetManager],
    ids=['trivial', 'zero_offset'])
@pytest.mark.parametrize('pack', [False, True], ids=['no_pack', 'pack'])
def test_tempalloc(cluda_api, tempalloc_cls, pack):

    shape = (10000,)
    dtype = numpy.int32
    thr = cluda_api.Thread.create(temp_alloc=dict(
        cls=tempalloc_cls, pack_on_alloc=False))

    # Dependency graph for the test
    dependencies = dict(
        _temp0=[],
        _temp1=['_temp9', '_temp8', '_temp3', '_temp5', '_temp4', '_temp7', '_temp6', 'input'],
        _temp10=['output', '_temp7'],
        _temp11=['_temp7'],
        _temp2=['input'],
        _temp3=['_temp1', 'input'],
        _temp4=['_temp9', '_temp8', '_temp1', '_temp7', '_temp6'],
        _temp5=['_temp1'],
        _temp6=['_temp1', '_temp4'],
        _temp7=['_temp9', '_temp1', '_temp4', 'output', '_temp11', '_temp10'],
        _temp8=['_temp1', '_temp4'],
        _temp9=['_temp1', '_temp4', '_temp7'],
        input=['_temp1', '_temp3', '_temp2'],
        output=['_temp10', '_temp7'])

    program = thr.compile(
    """
    KERNEL void fill(GLOBAL_MEM ${ctype} *dest, ${ctype} val)
    {
      const SIZE_T i = get_global_id(0);
      dest[i] = val;
    }

    KERNEL void transfer(GLOBAL_MEM ${ctype} *dest, GLOBAL_MEM ${ctype} *src)
    {
      const SIZE_T i = get_global_id(0);
      dest[i] = src[i];
    }
    """, render_kwds=dict(ctype=dtypes.ctype(dtype)))
    fill = program.fill
    transfer = program.transfer

    arrays = {}
    transfer_dest = thr.array(shape, dtype)

    # Allocate temporary arrays with dependencies
    for name in sorted(dependencies.keys()):
        deps = dependencies[name]
        arr_deps = [arrays[d] for d in deps if d in arrays]
        arrays[name] = thr.temp_array(shape, dtype, dependencies=arr_deps)
        fill(arrays[name], dtype(0), global_size=shape)

    if pack:
        thr.temp_alloc.pack()

    # Fill arrays with zeros
    for name in sorted(dependencies.keys()):
        deps = dependencies[name]
        arr_deps = [arrays[d] for d in deps if d in arrays]
        arrays[name] = thr.temp_array(shape, dtype, dependencies=arr_deps)
        fill(arrays[name], dtype(0), global_size=shape)

    for i, name in enumerate(sorted(dependencies.keys())):
        val = dtype(i + 1)
        fill(arrays[name], val, global_size=shape)
        for dep in dependencies[name]:
            # CUDA does not support get() for GPUArray with custom buffers,
            # So we need to transfer the data to a normal array first.
            transfer(transfer_dest, arrays[dep], global_size=shape)
            assert (transfer_dest.get() != val).all()


def test_constant_memory(thr):
    global_size = 100

    arr1 = get_test_array(global_size, numpy.int32)
    arr2 = get_test_array(global_size * 2, numpy.float32)
    ref = ((arr2[:global_size] + arr2[global_size:]) * arr1).astype(numpy.float32)

    src = """
        KERNEL void test(GLOBAL_MEM float *dest)
        {
            const SIZE_T i = get_global_id(0);
            dest[i] = (arr2[i] + arr2[i + get_global_size(0)]) * arr1[i];
        }
        """
    constant_arrays = dict(
        arr1=arr1, # use the parameters of an existing array; the array itself is not copied!
        arr2=(global_size * 2, numpy.float32)
        )

    # Compiled-in constant arrays only supported on CUDA
    if thr.api.get_id() != cluda.cuda_id():
        with pytest.raises(ValueError):
            program = thr.compile(src, constant_arrays=constant_arrays)
        return

    program = thr.compile(src, constant_arrays=constant_arrays)
    test = program.test

    # TODO: program.constant.arr1.set(arr1)
    program.set_constant('arr1', arr1)
    program.set_constant('arr2', arr2)

    dest_dev = thr.array(global_size, numpy.float32)
    test(dest_dev, global_size=global_size)

    assert diff_is_negligible(dest_dev.get(), ref)


    # Use a program/kernel with different constant arrays
    arr1_2 = get_test_array(global_size, numpy.int32)
    arr2_2 = get_test_array(global_size * 2, numpy.float32)
    ref_2 = ((arr2_2[:global_size] + arr2_2[global_size:]) * arr1_2).astype(numpy.float32)

    program2 = thr.compile(src, constant_arrays=constant_arrays)
    test2 = program2.test
    program2.set_constant('arr1', arr1_2)
    program2.set_constant('arr2', arr2_2)
    test2(dest_dev, global_size=global_size)
    assert diff_is_negligible(dest_dev.get(), ref_2)


    # Check that the first kernel preserved the data in its constant arrays
    test(dest_dev, global_size=global_size)
    assert diff_is_negligible(dest_dev.get(), ref)


def test_constant_memory_static_kernel(thr):
    global_size = 100

    arr1 = get_test_array(global_size, numpy.int32)
    arr2 = get_test_array(global_size * 2, numpy.float32)
    ref = ((arr2[:global_size] + arr2[global_size:]) * arr1).astype(numpy.float32)

    src = """
        KERNEL void test(GLOBAL_MEM float *dest)
        {
            const SIZE_T i = get_global_id(0);
            dest[i] = (arr2[i] + arr2[i + get_global_size(0)]) * arr1[i];
        }
        """
    constant_arrays = dict(
        arr1=arr1, # use the parameters of an existing array; the array itself is not copied!
        arr2=(global_size * 2, numpy.float32)
        )

    # Compiled-in constant arrays only supported on CUDA
    if thr.api.get_id() != cluda.cuda_id():
        with pytest.raises(ValueError):
            program = thr.compile_static(
                src, "test", global_size=global_size, constant_arrays=constant_arrays)
        return

    test = thr.compile_static(src, "test", global_size=global_size, constant_arrays=constant_arrays)

    # TODO: test.constant.arr1.set(arr1)
    test.set_constant('arr1', arr1)
    test.set_constant('arr2', thr.to_device(arr2)) # test that a device array is also accepted

    dest_dev = thr.array(global_size, numpy.float32)
    test(dest_dev)

    assert diff_is_negligible(dest_dev.get(), ref)


def test_offsets_in_kernel(thr):
    """
    Check that kernels receive the base data of arrays and have to add offsets manually.
    """

    global_size = 100
    dst_offset = 4
    src_offset = 2
    dtype = dtypes.normalize_type(numpy.int32)

    program = thr.compile("""
        KERNEL void test(GLOBAL_MEM int *dest, GLOBAL_MEM int *src)
        {
            const SIZE_T i = get_global_id(0);
            dest[i + ${dst_offset}] = src[i + ${src_offset}];
        }
        """,
        render_kwds=dict(dst_offset=dst_offset, src_offset=src_offset))
    test = program.test

    dest_dev_base = thr.array(global_size + dst_offset, dtype)
    dest_dev = thr.array(
        global_size, dtype, offset=dst_offset * dtype.itemsize, base=dest_dev_base)

    src_base = numpy.arange(global_size + src_offset).astype(dtype)
    src_dev_base = thr.to_device(src_base)
    src_dev = thr.array(global_size, dtype, offset=src_offset * dtype.itemsize, base=src_dev_base)

    test(dest_dev, src_dev, global_size=global_size)
    dest_ref = src_base[src_offset:]

    assert diff_is_negligible(dest_dev.get(), dest_ref)


class _GetSlices:
    def __getitem__(self, index):
        return index

_get_slices = _GetSlices()

_setitem_view_tests = [
    (_get_slices[:, :], 1, "[:,:]=scalar"),
    (_get_slices[:, 1], 1, "[:,i]=scalar"),
    (_get_slices[1, :], 1, "[i,:]=scalar"),
    (_get_slices[1, 2], 1, "[i,i]=scalar"),
    (_get_slices[:, :], numpy.arange(10 * 20).reshape(10, 20), "[:,:]=array"),
    (_get_slices[:, 2], numpy.arange(10), "[:,i]=array"),
    (_get_slices[2, :], numpy.arange(20), "[i,:]=array"),
]

@pytest.mark.parametrize(
    'setitem_test',
    [test[:2] for test in _setitem_view_tests],
    ids=[test[2] for test in _setitem_view_tests])
def test_setitem_view(thr, setitem_test):
    data = numpy.arange(10 * 20).reshape(10, 20).astype(numpy.int32)
    data_dev = thr.to_device(data)

    slices, value = setitem_test

    data[slices] = value
    data_dev[slices] = value

    assert diff_is_negligible(data_dev.get(), data)


# A regression test for a 0-dim array as a source for __setitem__
def test_setitem_0dim(thr):
    data = numpy.arange(10 * 20).reshape(10, 20).astype(numpy.int32)
    data_dev = thr.to_device(data)

    data[1, 2] = data[2, 3]
    data_dev[1, 2] = data_dev[2, 3]

    assert diff_is_negligible(data_dev.get(), data)


_get_view_tests = [
    (_get_slices[:, :], "[:,:]"),
    (_get_slices[:, 1], "[:,i]"),
    (_get_slices[1, :], "[i,:]"),
    (_get_slices[1, 2], "[i,i]"),
]

@pytest.mark.parametrize(
    'get_test_slices',
    [test[0] for test in _get_view_tests],
    ids=[test[1] for test in _get_view_tests])
def test_get_view(thr, get_test_slices):
    data = numpy.arange(10 * 20).reshape(10, 20).astype(numpy.int32)
    data_dev = thr.to_device(data)

    view_ref = data[get_test_slices]
    view_dev = data_dev[get_test_slices]

    assert diff_is_negligible(view_dev.get(), view_ref)


def test_concatenate(some_thr):

    template_shape = [10, None, 30]
    axis = 1
    dims = [3, 5, 10]

    arrays = []
    for i in range(len(dims)):
        template_shape[axis] = dims[i]
        arrays.append(get_test_array(tuple(template_shape), numpy.int32))

    ref = numpy.concatenate(arrays, axis=axis)

    arrays_dev = [some_thr.to_device(array) for array in arrays]
    test = concatenate(arrays_dev, axis=axis)

    assert diff_is_negligible(test.get(), ref)


@pytest.mark.parametrize('shift', [-10, 0, 11])
@pytest.mark.parametrize('axis', [0, 1, -1])
def test_roll(some_thr, shift, axis):

    array = get_test_array((5, 6, 7), numpy.int32)
    array_dev = some_thr.to_device(array)
    ref = numpy.roll(array, shift, axis=axis)
    test = roll(array_dev, shift, axis=axis)

    assert diff_is_negligible(test.get(), ref)


@pytest.mark.parametrize('shift', [-1, 0, 1])
@pytest.mark.parametrize('axis', [0, 1, -1])
def test_roll_method(some_thr, shift, axis):

    array = get_test_array((5, 6, 7), numpy.int32)
    array = numpy.arange(12).reshape(3, 4).astype(numpy.int32)
    array_dev = some_thr.to_device(array)
    ref = numpy.roll(array, shift, axis=axis)
    array_dev.roll(shift, axis=axis)

    assert diff_is_negligible(array_dev.get(), ref)



import time

import numpy
import pytest

from reikna.algorithms import Scan, Predicate, predicate_sum
from reikna.cluda import Snippet
import reikna.helpers as helpers
import reikna.cluda.dtypes as dtypes

from helpers import *


perf_shapes = [(1024 * 1024,), (1024 * 1024 * 8,), (1024 * 1024 * 64,)]
@pytest.fixture(params=perf_shapes, ids=list(map(str, perf_shapes)))
def large_perf_shape(request):
    return request.param


corr_shapes = [(15,), (511,), (512,), (513,), (512*512+1,), (512*512*4+5,)]
@pytest.fixture(params=corr_shapes, ids=list(map(str, corr_shapes)))
def corr_shape(request):
    return request.param


@pytest.fixture(params=[True, False], ids=["exclusive", "inclusive"])
def exclusive(request):
    return request.param


@pytest.fixture(params=[1, 2, 4, 8])
def seq_size(request):
    return request.param


def ref_scan(arr, axes=None, exclusive=False):
    if axes is None:
        res = numpy.cumsum(arr).reshape(arr.shape)
    else:
        axes = helpers.normalize_axes(arr.ndim, axes)
        transpose_to, transpose_from = helpers.make_axes_innermost(arr.ndim, axes)
        unchanged_ndim = arr.ndim - len(axes)
        temp = arr.transpose(transpose_to)
        temp2 = temp.reshape(
            temp.shape[:unchanged_ndim] + (helpers.product(temp.shape[unchanged_ndim:]),))
        temp2 = numpy.cumsum(temp2, axis=-1)
        res = temp2.reshape(temp.shape).transpose(transpose_from)

    if exclusive:
        res -= arr

    return res


def check_scan(
        thr, shape, axes, exclusive=False,
        measure_time=False, dtype=numpy.int64, max_work_group_size=None, predicate=None,
        seq_size=None):

    # Note: the comparison will only work if the custom predicate is
    # functionally equivalent to `predicate_sum`.
    if predicate is None:
        predicate = predicate_sum(dtype)

    arr = get_test_array(shape, dtype)

    scan = Scan(
        arr, predicate, axes=axes, exclusive=exclusive,
        max_work_group_size=max_work_group_size, seq_size=seq_size).compile(thr)

    arr_dev = thr.to_device(arr)
    res_dev = thr.to_device(numpy.ones_like(arr) * (-1))#thr.empty_like(arr)

    if measure_time:
        attempts = 10
        times = []
        for i in range(attempts):
            t1 = time.time()
            scan(res_dev, arr_dev)
            thr.synchronize()
            times.append(time.time() - t1)
        min_time = min(times)
    else:
        scan(res_dev, arr_dev)
        min_time = None

    res_test = res_dev.get()

    res_ref = ref_scan(arr, axes=axes, exclusive=exclusive)

    assert diff_is_negligible(res_ref, res_test)

    return min_time


def test_scan_correctness(thr, corr_shape, exclusive):
    check_scan(
        thr, corr_shape, axes=None, exclusive=exclusive,
        max_work_group_size=thr.device_params.max_work_group_size // 2)


def test_scan_multiple_axes(thr):
    check_scan(thr, (10, 20, 30, 40), axes=(2,3))


def test_scan_non_innermost_axes(thr):
    check_scan(thr, (10, 20, 30, 40), axes=(1,2))


def test_scan_custom_predicate(thr):
    predicate = Predicate(
        Snippet.create(lambda v1, v2: "return ${v1} + ${v2};"),
        0)
    check_scan(thr, (10, 20, 30, 40), axes=(1,2), predicate=predicate)


def test_scan_structure_type(thr, exclusive):

    shape = (100, 100)
    dtype = dtypes.align(numpy.dtype([
        ('i1', numpy.uint32),
        ('nested', numpy.dtype([
            ('v', numpy.uint64),
            ])),
        ('i2', numpy.uint32)
        ]))

    a = get_test_array(shape, dtype)
    a_dev = thr.to_device(a)

    # Have to construct the resulting array manually,
    # since numpy cannot scan arrays with struct dtypes.
    b_ref = numpy.empty(shape, dtype)
    b_ref['i1'] = ref_scan(a['i1'], axes=0, exclusive=exclusive)
    b_ref['nested']['v'] = ref_scan(a['nested']['v'], axes=0, exclusive=exclusive)
    b_ref['i2'] = ref_scan(a['i2'], axes=0, exclusive=exclusive)

    predicate = Predicate(
        Snippet.create(lambda v1, v2: """
            ${ctype} result = ${v1};
            result.i1 += ${v2}.i1;
            result.nested.v += ${v2}.nested.v;
            result.i2 += ${v2}.i2;
            return result;
            """,
            render_kwds=dict(
                ctype=dtypes.ctype_module(dtype))),
        numpy.zeros(1, dtype)[0])

    scan = Scan(a_dev, predicate, axes=(0,), exclusive=exclusive)

    b_dev = thr.empty_like(scan.parameter.output)

    scanc = scan.compile(thr)
    scanc(b_dev, a_dev)
    b_res = b_dev.get()

    assert diff_is_negligible(b_res, b_ref)


@pytest.mark.perf
@pytest.mark.returns('GB/s')
def test_large_scan_performance(thr, large_perf_shape, exclusive):
    """
    Large problem sizes.
    """
    dtype = dtypes.normalize_type(numpy.int64)
    min_time = check_scan(
        thr, large_perf_shape, dtype=dtype, axes=None, exclusive=exclusive, measure_time=True)
    return min_time, helpers.product(large_perf_shape) * dtype.itemsize


@pytest.mark.perf
@pytest.mark.returns('GB/s')
def test_small_scan_performance(thr, exclusive, seq_size):
    """
    Small problem sizes, big batches.
    """
    dtype = dtypes.normalize_type(numpy.complex128)
    shape = (500, 2, 2, 256)
    min_time = check_scan(
        thr, shape, dtype=dtype, axes=(-1,), exclusive=exclusive,
        measure_time=True, seq_size=seq_size)
    return min_time, helpers.product(shape) * dtype.itemsize



import pytest

from reikna.helpers import template_from
from reikna.core import Computation, Parameter, Annotation, Transformation, Type

from helpers import *


class Dummy(Computation):
    """
    Dummy computation class with two inputs, two outputs and one parameter.
    Used to perform core and transformation tests.
    """

    def __init__(self, length, arr1, arr2):
        assert arr1.shape == (length,)
        assert arr2.shape == (2, length)
        self._arr1 = arr1
        self._arr2 = arr2
        Computation.__init__(self, [
            Parameter('output', Annotation(Type(numpy.float32, length), 'o')),
            ])

    def _build_plan(self, plan_factory, device_params, output):
        plan = plan_factory()

        template = template_from("""
        <%def name="dummy(kernel_declaration, output, arr1, arr2)">
        ${kernel_declaration}
        {
            VIRTUAL_SKIP_THREADS;
            const VSIZE_T i = virtual_global_id(0);
            ${arr1.ctype} x1 = ${arr1.load_idx}(i);
            ${arr2.ctype} x2 = ${arr2.load_idx}(0, i);
            ${arr2.ctype} x3 = ${arr2.load_idx}(1, i);
            ${output.store_idx}(i, (x2 + x3) * x1);
        }
        </%def>
        """)

        arr1 = plan.constant_array(self._arr1)
        arr2 = plan.constant_array(self._arr2)

        plan.kernel_call(
            template.get_def('dummy'),
            [output, arr1, arr2],
            global_size=output.shape)

        return plan


class DummyOuter(Computation):

    def __init__(self, length, arr1, arr2):
        assert arr1.shape == (length,)
        assert arr2.shape == (2, length)
        self._arr1 = arr1
        self._arr2 = arr2
        Computation.__init__(self, [
            Parameter('output', Annotation(Type(numpy.float32, length), 'o')),
            ])

    def _build_plan(self, plan_factory, device_params, output):
        plan = plan_factory()
        dummy = Dummy(self._arr1.shape[0], self._arr1, self._arr2)
        plan.computation_call(dummy, output)
        return plan


def test_constant_arrays_computation(thr):

    N = 200
    arr1 = get_test_array(N, numpy.int32)
    arr2 = get_test_array((2, N), numpy.float32)
    ref = (arr1 * (arr2[0] + arr2[1])).astype(numpy.float32)

    d = Dummy(N, arr1, arr2).compile(thr)
    out_dev = thr.empty_like(d.parameter.output)
    d(out_dev)
    test = out_dev.get()

    assert diff_is_negligible(test, ref)


def test_constant_arrays_computation_nested(thr):
    """
    Check that constant arrays from a nested computation are
    transfered to the outer computation.
    """

    N = 200
    arr1 = get_test_array(N, numpy.int32)
    arr2 = get_test_array((2, N), numpy.float32)
    ref = (arr1 * (arr2[0] + arr2[1])).astype(numpy.float32)

    d = DummyOuter(N, arr1, arr2).compile(thr)
    out_dev = thr.empty_like(d.parameter.output)
    d(out_dev)
    test = out_dev.get()

    assert diff_is_negligible(test, ref)



"""
This example demonstrates how to build several specialized FFT computations:
- a real-to-complex and complex-to-real FFT using a half-sized complex-to-complex FFT;
- a real-to-complex and complex-to-real FFT with the real signal being anti-periodic
  (x[k] = -x[N/2+k], k=0..N/2-1) using a quarter-sized complex-to-complex FFT.

Note that in the second case all the harmonics with even (0-based) indices are equal to zero,
so both the reference and the GPU FFT return only N//4 odd harmonics
(and, correspondingly, the inverse functions take an N//4-array of odd harmonics
and return the first half of the initial array).

For the sake of the example only even N and 1D FFT are supported.

Source of the algorithms:
L. R. Rabiner, "On the Use of Symmetry in FFT Computation"
IEEE Transactions on Acoustics, Speech, and Signal Processing 27(3), 233-239 (1979)
doi: 10.1109/TASSP.1979.1163235
"""

import numpy

from reikna.fft import FFT

import reikna.helpers as helpers
from reikna.cluda import dtypes, any_api
from reikna.core import Computation, Parameter, Annotation, Type, Transformation
import reikna.cluda.functions as functions
from reikna.algorithms import Reduce, Scan, predicate_sum

TEMPLATE = helpers.template_for(__file__)


# Refernce functions


def rfft_reference(a):
    # Numpy already has a dedicated rfft() function, but this one illustrates the algorithm
    # used by the computation.

    assert a.size % 2 == 0
    assert a.ndim == 1

    N = a.size

    WNmk = numpy.exp(-2j * numpy.pi * numpy.arange(N//2) / N)
    A = 0.5 * (1 - 1j * WNmk)
    B = 0.5 * (1 + 1j * WNmk)

    x = a[::2] + 1j * a[1::2]
    X = numpy.fft.fft(x) # N/2-sized FFT

    res = numpy.empty(N//2 + 1, numpy.complex128)
    res[:N//2] = X * A + (numpy.roll(X[N//2-1::-1], 1)).conj() * B
    res[N//2] = X[0].real - X[0].imag

    return res


def irfft_reference(a):
    # Numpy already has a dedicated irfft() function, but this one illustrates the algorithm
    # used by the computation.

    assert a.size % 2 == 1
    assert a.ndim == 1

    N = (a.size - 1) * 2

    # Following numpy.fft.irftt() which ignores these values
    a[0] = a[0].real
    a[-1] = a[-1].real

    WNmk = numpy.exp(-2j * numpy.pi * numpy.arange(N//2) / N)
    A = 0.5 * (1 - 1j * WNmk)
    B = 0.5 * (1 + 1j * WNmk)

    A = A.conj()
    B = B.conj()

    X = a[:-1] * A + a[N//2:0:-1].conj() * B

    x = numpy.fft.ifft(X) # N/2-sized IFFT

    res = numpy.empty(N, numpy.float64)
    res[::2] = x.real
    res[1::2] = x.imag

    return res


def aprfft_reference(x):
    # x : real N/2-array (first half of the full signal)
    # output: complex N/4-array (odd harmonics)

    assert x.size % 2 == 0
    assert x.ndim == 1

    N = x.size * 2

    y = x * (4 * numpy.sin(2 * numpy.pi * numpy.arange(N//2) / N))
    Y = numpy.fft.rfft(y)

    t = x * numpy.cos(2 * numpy.pi * numpy.arange(N//2) / N)
    re_X_1 = 2 * sum(t)

    Y *= -1j
    Y[0] /= 2
    Y[0] += re_X_1
    res = numpy.cumsum(Y[:-1])

    return res


def iaprfft_reference(X):
    # X : complex N/4-array of odd harmonics
    # output: real N/2-array (first half of the signal)

    assert X.ndim == 1

    N = X.size * 4
    Y = numpy.empty(N//4+1, numpy.complex128)
    Y[1:-1] = X[1:] - X[:-1]
    Y *= 1j
    Y[0] = -2 * X[0].imag
    Y[-1] = 2 * X[-1].imag
    y = numpy.fft.irfft(Y)

    res = numpy.empty(N//2, numpy.float64)
    res[1:] = y[1:] / 4 / numpy.sin(2 * numpy.pi * numpy.arange(1, N//2) / N)
    res[0] = X.real.sum() / (N / 2)

    return res


# GPU computations


def prepare_rfft_input(arr):
    res = Type(dtypes.complex_for(arr.dtype), arr.shape[:-1] + (arr.shape[-1] // 2,))
    return Transformation(
        [
            Parameter('output', Annotation(res, 'o')),
            Parameter('input', Annotation(arr, 'i')),
        ],
        """
        <%
            batch_idxs = " ".join((idx + ", ") for idx in idxs[:-1])
        %>
        ${input.ctype} re = ${input.load_idx}(${batch_idxs} ${idxs[-1]} * 2);
        ${input.ctype} im = ${input.load_idx}(${batch_idxs} ${idxs[-1]} * 2 + 1);
        ${output.store_same}(COMPLEX_CTR(${output.ctype})(re, im));
        """,
        connectors=['output'])


class RFFT(Computation):

    def __init__(self, arr_t, dont_store_last=False):
        self._dont_store_last = dont_store_last

        output_size = arr_t.shape[-1] // 2 + (0 if dont_store_last else 1)

        out_arr = Type(
            dtypes.complex_for(arr_t.dtype),
            arr_t.shape[:-1] + (output_size,))

        Computation.__init__(self, [
            Parameter('output', Annotation(out_arr, 'o')),
            Parameter('input', Annotation(arr_t, 'i'))])

    def _build_plan(self, plan_factory, device_params, output, input_):

        plan = plan_factory()

        N = input_.shape[-1]
        WNmk = numpy.exp(-2j * numpy.pi * numpy.arange(N//2) / N)
        A = 0.5 * (1 - 1j * WNmk)
        B = 0.5 * (1 + 1j * WNmk)

        A_arr = plan.persistent_array(A)
        B_arr = plan.persistent_array(B)

        cfft_arr = Type(output.dtype, input_.shape[:-1] + (input_.shape[-1] // 2,))
        cfft = FFT(cfft_arr, axes=(len(input_.shape) - 1,))

        prepare_input = prepare_rfft_input(input_)

        cfft.parameter.input.connect(
            prepare_input, prepare_input.output, real_input=prepare_input.input)

        temp = plan.temp_array_like(cfft.parameter.output)

        batch_size = helpers.product(output.shape[:-1])

        plan.computation_call(cfft, temp, input_)
        plan.kernel_call(
            TEMPLATE.get_def('prepare_rfft_output'),
                [output, temp, A_arr, B_arr],
                global_size=(batch_size, N // 2),
                render_kwds=dict(
                    slices=(len(input_.shape) - 1, 1),
                    N=N,
                    mul=functions.mul(output.dtype, output.dtype),
                    conj=functions.conj(output.dtype),
                    dont_store_last=self._dont_store_last))

        return plan


def prepare_irfft_output(arr):
    res = Type(dtypes.real_for(arr.dtype), arr.shape[:-1] + (arr.shape[-1] * 2,))
    return Transformation(
        [
            Parameter('output', Annotation(res, 'o')),
            Parameter('input', Annotation(arr, 'i')),
        ],
        """
        <%
            batch_idxs = " ".join((idx + ", ") for idx in idxs[:-1])
        %>
        ${input.ctype} x = ${input.load_same};
        ${output.store_idx}(${batch_idxs} ${idxs[-1]} * 2, x.x);
        ${output.store_idx}(${batch_idxs} ${idxs[-1]} * 2 + 1, x.y);
        """,
        connectors=['output'])


class IRFFT(Computation):

    def __init__(self, arr_t):

        output_size = (arr_t.shape[-1] - 1) * 2

        out_arr = Type(
            dtypes.real_for(arr_t.dtype),
            arr_t.shape[:-1] + (output_size,))

        Computation.__init__(self, [
            Parameter('output', Annotation(out_arr, 'o')),
            Parameter('input', Annotation(arr_t, 'i'))])

    def _build_plan(self, plan_factory, device_params, output, input_):

        plan = plan_factory()

        N = (input_.shape[-1] - 1) * 2

        WNmk = numpy.exp(-2j * numpy.pi * numpy.arange(N//2) / N)
        A = 0.5 * (1 - 1j * WNmk)
        B = 0.5 * (1 + 1j * WNmk)

        A_arr = plan.persistent_array(A.conj())
        B_arr = plan.persistent_array(B.conj())

        cfft_arr = Type(input_.dtype, input_.shape[:-1] + (N // 2,))
        cfft = FFT(cfft_arr, axes=(len(input_.shape) - 1,))

        prepare_output = prepare_irfft_output(cfft.parameter.output)

        cfft.parameter.output.connect(
            prepare_output, prepare_output.input, real_output=prepare_output.output)

        temp = plan.temp_array_like(cfft.parameter.input)

        batch_size = helpers.product(output.shape[:-1])

        plan.kernel_call(
            TEMPLATE.get_def('prepare_irfft_input'),
                [temp, input_, A_arr, B_arr],
                global_size=(batch_size, N // 2),
                render_kwds=dict(
                    slices=(len(input_.shape) - 1, 1),
                    N=N,
                    mul=functions.mul(input_.dtype, input_.dtype),
                    conj=functions.conj(input_.dtype)))

        plan.computation_call(cfft, output, temp, inverse=True)

        return plan


def get_multiply(output):
    return Transformation(
        [
            Parameter('output', Annotation(output, 'o')),
            Parameter('a', Annotation(output, 'i')),
            Parameter('b', Annotation(Type(output.dtype, (output.shape[-1],)), 'i'))
        ],
        """
        ${output.store_same}(${mul}(${a.load_same}, ${b.load_idx}(${idxs[-1]})));
        """,
        connectors=['output', 'a'],
        render_kwds=dict(mul=functions.mul(output.dtype, output.dtype))
        )


def get_prepare_prfft_scan(output):
    return Transformation(
        [
            Parameter('output', Annotation(output, 'o')),
            Parameter('Y', Annotation(output, 'i')),
            Parameter('re_X_0', Annotation(
                Type(dtypes.real_for(output.dtype), output.shape[:-1]), 'i'))
        ],
        """
        ${Y.ctype} Y = ${Y.load_same};
        Y = COMPLEX_CTR(${Y.ctype})(Y.y, -Y.x);

        if (${idxs[-1]} == 0)
        {
            Y.x = Y.x / 2 + ${re_X_0.load_idx}(${", ".join(idxs[:-1])});
            Y.y /= 2;
        }

        ${output.store_same}(Y);
        """,
        connectors=['output', 'Y'],
        )


class APRFFT(Computation):
    """
    FFT of a real antiperiodic signal (x[k] = -x[N/2+k]).
    """

    def __init__(self, arr_t):

        out_arr = Type(
            dtypes.complex_for(arr_t.dtype),
            arr_t.shape[:-1] + (arr_t.shape[-1] // 2,))

        Computation.__init__(self, [
            Parameter('output', Annotation(out_arr, 'o')),
            Parameter('input', Annotation(arr_t, 'i'))])

    def _build_plan(self, plan_factory, device_params, output, input_):

        plan = plan_factory()

        N = input_.shape[-1] * 2
        batch_shape = input_.shape[:-1]
        batch_size = helpers.product(batch_shape)

        coeffs1 = 4 * numpy.sin(2 * numpy.pi * numpy.arange(N//2) / N)
        coeffs2 = 2 * numpy.cos(2 * numpy.pi * numpy.arange(N//2) / N)

        c1_arr = plan.persistent_array(coeffs1)
        c2_arr = plan.persistent_array(coeffs2)

        multiply = get_multiply(input_)

        # re_X_1 = sum(x * coeffs2)

        t = plan.temp_array_like(input_)
        rd = Reduce(t, predicate_sum(input_.dtype), axes=(len(input_.shape)-1,))

        rd.parameter.input.connect(
            multiply, multiply.output, x=multiply.a, c2=multiply.b)

        re_X_0 = plan.temp_array_like(rd.parameter.output)
        plan.computation_call(rd, re_X_0, input_, c2_arr)

        # Y = numpy.fft.rfft(x * coeffs1)

        rfft = RFFT(input_, dont_store_last=True)
        rfft.parameter.input.connect(
            multiply, multiply.output, x=multiply.a, c1=multiply.b)

        Y = plan.temp_array_like(rfft.parameter.output)
        plan.computation_call(rfft, Y, input_, c1_arr)

        # Y *= -1j
        # Y[0] /= 2
        # Y[0] += re_X_1
        # res = numpy.cumsum(Y[:-1])

        prepare_prfft_scan = get_prepare_prfft_scan(Y)

        sc = Scan(Y, predicate_sum(Y.dtype), axes=(-1,), exclusive=False)
        sc.parameter.input.connect(
            prepare_prfft_scan, prepare_prfft_scan.output,
            Y=prepare_prfft_scan.Y, re_X_0=prepare_prfft_scan.re_X_0)

        plan.computation_call(sc, output, Y, re_X_0)

        return plan


def get_prepare_iprfft_input(X):
    # Input: size N//4
    # Output: size N//4+1

    N = X.shape[-1] * 4
    Y = Type(X.dtype, X.shape[:-1] + (N // 4 + 1,))

    return Transformation(
        [
            Parameter('Y', Annotation(Y, 'o')),
            Parameter('X', Annotation(X, 'i')),
        ],
        """
        <%
            batch_idxs = " ".join((idx + ", ") for idx in idxs[:-1])
        %>

        ${Y.ctype} Y;
        if (${idxs[-1]} == 0)
        {
            ${X.ctype} X = ${X.load_idx}(${batch_idxs} 0);
            Y = COMPLEX_CTR(${Y.ctype})(-2 * X.y, 0);
        }
        else if (${idxs[-1]} == ${N//4})
        {
            ${X.ctype} X = ${X.load_idx}(${batch_idxs} ${N//4-1});
            Y = COMPLEX_CTR(${Y.ctype})(2 * X.y, 0);
        }
        else
        {
            ${X.ctype} X = ${X.load_idx}(${batch_idxs} ${idxs[-1]});
            ${X.ctype} X_prev = ${X.load_idx}(${batch_idxs} ${idxs[-1]} - 1);
            ${X.ctype} diff = X - X_prev;
            Y = COMPLEX_CTR(${Y.ctype})(-diff.y, diff.x);
        }

        ${Y.store_same}(Y);
        """,
        connectors=['Y'],
        render_kwds=dict(N=N)
        )


def get_prepare_iprfft_output(y):
    # Input: size N//4
    # Output: size N//4

    N = y.shape[-1] * 2

    return Transformation(
        [
            Parameter('x', Annotation(y, 'o')),
            Parameter('y', Annotation(y, 'i')),
            Parameter('x0', Annotation(Type(y.dtype, y.shape[:-1]), 'i')),
            Parameter('coeffs', Annotation(Type(y.dtype, (N//2,)), 'i')),
        ],
        """
        ${y.ctype} y = ${y.load_same};
        ${coeffs.ctype} coeff = ${coeffs.load_idx}(${idxs[-1]});

        ${x.ctype} x;

        if (${idxs[-1]} == 0)
        {
            ${x0.ctype} x0 = ${x0.load_idx}(${", ".join(idxs[:-1])});
            x = x0 / ${N // 2};
        }
        else
        {
            x = y * coeff;
        }

        ${x.store_same}(x);
        """,
        connectors=['y'],
        render_kwds=dict(N=N)
        )


class IAPRFFT(Computation):
    """
    IFFT of a real antiperiodic signal (x[k] = -x[N/2+k]).
    """

    def __init__(self, arr_t):

        out_arr = Type(
            dtypes.real_for(arr_t.dtype),
            arr_t.shape[:-1] + (arr_t.shape[-1] * 2,))

        Computation.__init__(self, [
            Parameter('output', Annotation(out_arr, 'o')),
            Parameter('input', Annotation(arr_t, 'i'))])

    def _build_plan(self, plan_factory, device_params, output, input_):

        plan = plan_factory()

        N = input_.shape[-1] * 4
        batch_shape = input_.shape[:-1]
        batch_size = helpers.product(batch_shape)

        # The first element is unused
        coeffs = numpy.concatenate(
            [[0], 1 / (4 * numpy.sin(2 * numpy.pi * numpy.arange(1, N//2) / N))])
        coeffs_arr = plan.persistent_array(coeffs)

        prepare_iprfft_input = get_prepare_iprfft_input(input_)
        prepare_iprfft_output = get_prepare_iprfft_output(output)

        irfft = IRFFT(prepare_iprfft_input.Y)
        irfft.parameter.input.connect(
            prepare_iprfft_input, prepare_iprfft_input.Y,
            X=prepare_iprfft_input.X)
        irfft.parameter.output.connect(
            prepare_iprfft_output, prepare_iprfft_output.y,
            x=prepare_iprfft_output.x,
            x0=prepare_iprfft_output.x0, coeffs=prepare_iprfft_output.coeffs)

        real = Transformation(
            [
                Parameter('output', Annotation(Type(dtypes.real_for(input_.dtype), input_.shape), 'o')),
                Parameter('input', Annotation(input_, 'i')),
            ],
            """
            ${output.store_same}((${input.load_same}).x);
            """,
            connectors=['output']
            )

        rd_t = Type(output.dtype, input_.shape)
        rd = Reduce(rd_t, predicate_sum(rd_t.dtype), axes=(len(input_.shape)-1,))
        rd.parameter.input.connect(real, real.output, X=real.input)

        x0 = plan.temp_array_like(rd.parameter.output)

        plan.computation_call(rd, x0, input_)
        plan.computation_call(irfft, output, x0, coeffs_arr, input_)

        return plan


# Tests


def test_rfft(thr):
    N = 1024
    a = numpy.random.normal(size=N)

    rfft = RFFT(a).compile(thr)

    fa_numpy = numpy.fft.rfft(a)
    fa_ref = rfft_reference(a)

    a_dev = thr.to_device(a)
    fa_gpu = thr.empty_like(rfft.parameter.output)
    rfft(fa_gpu, a_dev)

    assert numpy.allclose(fa_numpy, fa_ref)
    assert numpy.allclose(fa_numpy, fa_gpu.get())


def test_irfft(thr):
    N = 1024
    fa = (numpy.random.normal(size=N//2+1) + 1j * numpy.random.normal(size=N//2+1))

    irfft = IRFFT(fa).compile(thr)

    a_numpy = numpy.fft.irfft(fa)
    a_ref = irfft_reference(fa)

    fa_dev = thr.to_device(fa)
    a_gpu = thr.empty_like(irfft.parameter.output)
    irfft(a_gpu, fa_dev)

    assert numpy.allclose(a_numpy, a_ref)
    assert numpy.allclose(a_numpy, a_gpu.get())


def test_aprfft(thr):
    N = 1024
    half_a = numpy.random.normal(size=N//2)
    a = numpy.concatenate([half_a, -half_a])

    aprfft = APRFFT(half_a).compile(thr)

    fa_numpy = numpy.fft.rfft(a)[1::2]
    fa_ref = aprfft_reference(half_a)

    half_a_dev = thr.to_device(half_a)
    fa_dev = thr.empty_like(aprfft.parameter.output)
    aprfft(fa_dev, half_a_dev)

    assert numpy.allclose(fa_numpy, fa_ref)
    assert numpy.allclose(fa_numpy, fa_dev.get())


def test_iaprfft(thr):
    N = 1024
    fa_odd_harmonics = (numpy.random.normal(size=N//4) + 1j * numpy.random.normal(size=N//4))

    fa = numpy.zeros(N//2+1, fa_odd_harmonics.dtype)
    fa[1::2] = fa_odd_harmonics

    iaprfft = IAPRFFT(fa_odd_harmonics).compile(thr)

    half_a_numpy = numpy.fft.irfft(fa)[:N//2]
    half_a_ref = iaprfft_reference(fa_odd_harmonics)

    fa_oh_dev = thr.to_device(fa_odd_harmonics)
    half_a_dev = thr.empty_like(iaprfft.parameter.output)
    iaprfft(half_a_dev, fa_oh_dev)

    assert numpy.allclose(half_a_numpy, half_a_ref)
    assert numpy.allclose(half_a_numpy, half_a_dev.get())


if __name__ == '__main__':
    api = any_api()
    thr = api.Thread.create(interactive=True)

    test_rfft(thr)
    test_irfft(thr)
    test_aprfft(thr)
    test_iaprfft(thr)



"""
This example demonstrates how to write an analogue of the ``matplotlib.mlab.specgram``
function using Reikna. The main difficulty here is to split the task into transformations
and computation cores and assemble them into a single computation.

To map the task to the transformation/computation model, one needs to separate the steps
into those that do not require the communication between threads, and those that do.
The former will be transformations, the latter will be computations.

``specgram`` does the following:

* Reshapes the initial 1D array into a 2D with an overlap between rows;
* Applies a window function;
* Performs an FFT over the rows;
* Crops the negative frequencies part of the result (default behavior);
* Calculates the spectrum. The actual formula here depends on the ``mode`` parameter.
  This example uses ``mode='amplitude'`` to make the example simpler,
  because the default ``mode='psd'`` does a bit of additional scaling.
  Implementing other modes is left as an exercise for the reader.

Of all these, only FFT is a computation, the rest can be implemented as transformations.

Note that ``specgram`` returns the array of shape ``(frequencies, times)``.
Reikna's batched FFT is most effective when performed over the last axis of the array,
so we will operate with ``(times, frequencies)`` array and transpose it in the end.
Transposition is also a computation, so we will have to chain it after the FFT.

Since the FFT requires a complex-valued array, we will need an additional transformation that
typecasts the initial real-valued array into complex numbers. One could write a custom
computation that does that, but we will use two predefined transformations
(from the ``reikna.transformations`` module): ``combine_complex`` (creates a complex array
out of real and imaginary parts) and ``broadcast_const`` (to fill the complex part with zeroes).

CAVEAT: In the steps above, two of the transformations (the initial reshape and the frequency crop)
will change the shape of the array. Currently such transformations can only be leaves of the
transformation tree, so we will have to swap the cropping step and the spectrum calculating step.
"""

import numpy

import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
from matplotlib.mlab import specgram, window_hanning

from reikna.cluda import any_api
from reikna.cluda import dtypes, functions
from reikna.core import Computation, Transformation, Parameter, Annotation, Type
from reikna.fft import FFT
from reikna.algorithms import Transpose
import reikna.transformations as transformations


def get_data():

    dt = 0.0005
    t = numpy.arange(0.0, 20.0, dt)
    s1 = numpy.sin(2*numpy.pi*100*t)
    s2 = 2*numpy.sin(2*numpy.pi*400*t)

    # create a transient "chirp"
    mask = numpy.where(numpy.logical_and(t > 10, t < 12), 1.0, 0.0)
    s2 = s2 * mask

    # add some noise into the mix
    nse = 0.01*numpy.random.randn(len(t))

    x = s1 + s2 + nse # the signal
    NFFT = 1024       # the length of the windowing segments
    Fs = int(1.0/dt)  # the sampling frequency

    return x, dict(NFFT=NFFT, Fs=Fs, noverlap=900, pad_to=2048)


def hanning_window(arr, NFFT):
    """
    Applies the von Hann window to the rows of a 2D array.
    To account for zero padding (which we do not want to window), NFFT is provided separately.
    """
    if dtypes.is_complex(arr.dtype):
        coeff_dtype = dtypes.real_for(arr.dtype)
    else:
        coeff_dtype = arr.dtype
    return Transformation(
        [
            Parameter('output', Annotation(arr, 'o')),
            Parameter('input', Annotation(arr, 'i')),
        ],
        """
        ${dtypes.ctype(coeff_dtype)} coeff;
        %if NFFT != output.shape[0]:
        if (${idxs[1]} >= ${NFFT})
        {
            coeff = 1;
        }
        else
        %endif
        {
            coeff = 0.5 * (1 - cos(2 * ${numpy.pi} * ${idxs[-1]} / (${NFFT} - 1)));
        }
        ${output.store_same}(${mul}(${input.load_same}, coeff));
        """,
        render_kwds=dict(
            coeff_dtype=coeff_dtype, NFFT=NFFT,
            mul=functions.mul(arr.dtype, coeff_dtype)))


def rolling_frame(arr, NFFT, noverlap, pad_to):
    """
    Transforms a 1D array to a 2D array whose rows are
    partially overlapped parts of the initial array.
    """

    frame_step = NFFT - noverlap
    frame_num = (arr.size - noverlap) // frame_step
    frame_size = NFFT if pad_to is None else pad_to

    result_arr = Type(arr.dtype, (frame_num, frame_size))

    return Transformation(
        [
            Parameter('output', Annotation(result_arr, 'o')),
            Parameter('input', Annotation(arr, 'i')),
        ],
        """
        %if NFFT != output.shape[1]:
        if (${idxs[1]} >= ${NFFT})
        {
            ${output.store_same}(0);
        }
        else
        %endif
        {
            ${output.store_same}(${input.load_idx}(${idxs[0]} * ${frame_step} + ${idxs[1]}));
        }
        """,
        render_kwds=dict(frame_step=frame_step, NFFT=NFFT),
        # note that only the "store_same"-using argument can serve as a connector!
        connectors=['output'])


def crop_frequencies(arr):
    """
    Crop a 2D array whose columns represent frequencies to only leave the frequencies with
    different absolute values.
    """
    result_arr = Type(arr.dtype, (arr.shape[0], arr.shape[1] // 2 + 1))
    return Transformation(
        [
            Parameter('output', Annotation(result_arr, 'o')),
            Parameter('input', Annotation(arr, 'i')),
        ],
        """
        if (${idxs[1]} < ${input.shape[1] // 2 + 1})
            ${output.store_idx}(${idxs[0]}, ${idxs[1]}, ${input.load_same});
        """,
        # note that only the "load_same"-using argument can serve as a connector!
        connectors=['input'])


class Spectrogram(Computation):

    def __init__(self, x, NFFT=256, noverlap=128, pad_to=None, window=hanning_window):

        assert dtypes.is_real(x.dtype)
        assert x.ndim == 1

        rolling_frame_trf = rolling_frame(x, NFFT, noverlap, pad_to)

        complex_dtype = dtypes.complex_for(x.dtype)
        fft_arr = Type(complex_dtype, rolling_frame_trf.output.shape)
        real_fft_arr = Type(x.dtype, rolling_frame_trf.output.shape)

        window_trf = window(real_fft_arr, NFFT)
        broadcast_zero_trf = transformations.broadcast_const(real_fft_arr, 0)
        to_complex_trf = transformations.combine_complex(fft_arr)
        amplitude_trf = transformations.norm_const(fft_arr, 1)
        crop_trf = crop_frequencies(amplitude_trf.output)

        fft = FFT(fft_arr, axes=(1,))
        fft.parameter.input.connect(
            to_complex_trf, to_complex_trf.output,
            input_real=to_complex_trf.real, input_imag=to_complex_trf.imag)
        fft.parameter.input_imag.connect(
            broadcast_zero_trf, broadcast_zero_trf.output)
        fft.parameter.input_real.connect(
            window_trf, window_trf.output, unwindowed_input=window_trf.input)
        fft.parameter.unwindowed_input.connect(
            rolling_frame_trf, rolling_frame_trf.output, flat_input=rolling_frame_trf.input)
        fft.parameter.output.connect(
            amplitude_trf, amplitude_trf.input, amplitude=amplitude_trf.output)
        fft.parameter.amplitude.connect(
            crop_trf, crop_trf.input, cropped_amplitude=crop_trf.output)

        self._fft = fft

        self._transpose = Transpose(fft.parameter.cropped_amplitude)

        Computation.__init__(self,
            [Parameter('output', Annotation(self._transpose.parameter.output, 'o')),
            Parameter('input', Annotation(fft.parameter.flat_input, 'i'))])

    def _build_plan(self, plan_factory, device_params, output, input_):
        plan = plan_factory()
        temp = plan.temp_array_like(self._fft.parameter.cropped_amplitude)
        plan.computation_call(self._fft, temp, input_)
        plan.computation_call(self._transpose, output, temp)
        return plan


if __name__ == '__main__':

    numpy.random.seed(125)
    x, params = get_data()

    fig = plt.figure()
    s = fig.add_subplot(2, 1, 1)
    spectre, freqs, ts = specgram(x, mode='magnitude', **params)

    # Renormalize to match the computation
    spectre *= window_hanning(numpy.ones(params['NFFT'])).sum()

    s.imshow(
        numpy.log10(spectre),
        extent=(ts[0], ts[-1], freqs[0], freqs[-1]),
        aspect='auto',
        origin='lower')

    api = any_api()
    thr = api.Thread.create()

    specgram_reikna = Spectrogram(
        x, NFFT=params['NFFT'], noverlap=params['noverlap'], pad_to=params['pad_to']).compile(thr)

    x_dev = thr.to_device(x)
    spectre_dev = thr.empty_like(specgram_reikna.parameter.output)
    specgram_reikna(spectre_dev, x_dev)
    spectre_reikna = spectre_dev.get()

    assert numpy.allclose(spectre, spectre_reikna)

    s = fig.add_subplot(2, 1, 2)
    im=s.imshow(
        numpy.log10(spectre_reikna),
        extent=(ts[0], ts[-1], freqs[0], freqs[-1]),
        aspect='auto',
        origin='lower')
    fig.savefig('demo_specgram.png')
